package com.laguerre.kan;

import java.nio.FloatBuffer;

import android.opengl.GLES20;
import android.util.Log;

public final class PassShader {
	public static enum CullFace {
		BACK,
		FRONT,
		FRONT_AND_BACK
	}

	public static enum BlendSrc {
		ZERO,
		ONE,
		COLOR,
		ONE_MINUS_COLOR,
		ALPHA,
		ONE_MINUS_ALPHA,
		CONSTANT_COLOR,
		ONE_MINUS_CONSTANT_COLOR,
		CONSTANT_ALPHA,
		ONE_MINUS_CONSTANT_ALPHA,
		ALPHA_SATURATE
	}

	public static enum BlendDst {
		ZERO,
		ONE,
		COLOR,
		ONE_MINUS_COLOR,
		ALPHA,
		ONE_MINUS_ALPHA,
		CONSTANT_COLOR,
		ONE_MINUS_CONSTANT_COLOR,
		CONSTANT_ALPHA,
		ONE_MINUS_CONSTANT_ALPHA
	}

	public static enum DepthTest {
		LESS,
		GREATER,
		LEQUAL,
		GEQUAL,
		EQUAL,
		NOT_EQUAL,
		ALWAYS
	}

	private int programHandle = -1;
	private boolean doBackfaceCulling;
	private CullFace cullFace;
	private boolean doBlending;
	private BlendSrc blendSrc;
	private BlendDst blendDst;
	private final Color constantBlendColor = Color.getNewWhite();
	private final boolean[] colorMask = { true, true, true, true };
	private boolean doDepthWriting;
	private DepthTest depthTest = DepthTest.ALWAYS;

	/**
	 * See <code>Shader.addPassShader</code> to create a {@link PassShader}.
	 */
	PassShader() {
		setCullBackface(true);
		setCullFace(CullFace.BACK);
		setBlending(false);
		setBlendFunc(BlendSrc.ONE, BlendDst.ONE);
		setDepthWriting(true);
	}

	public void setCullBackface(boolean doBackfaceCulling) {
		this.doBackfaceCulling = doBackfaceCulling;
	}

	public void setBlending(boolean doBlending) {
		this.doBlending = doBlending;
	}

	public void setConstantBlendColor(Color c) {
		this.constantBlendColor.setValues(c);
	}

	public void setConstantBlendColor(float r, float g, float b, float a) {
		this.constantBlendColor.setValues(r, g, b, a);
	}

	public void setBlendFunc(BlendSrc blendSrc, BlendDst blendDst) {
		this.blendSrc = blendSrc;
		this.blendDst = blendDst;
	}

	private int getCullFaceParam(CullFace cullFace) {
		int ret = -1;
		switch (cullFace) {
		case BACK:
			ret = GLES20.GL_BACK;
			break;
		case FRONT:
			ret = GLES20.GL_FRONT;
			break;
		case FRONT_AND_BACK:
			ret = GLES20.GL_FRONT_AND_BACK;
			break;
		}
		return ret;
	}

	public void setCullFace(CullFace cullFace) {
		this.cullFace = cullFace;
	}

	private int getBlenSrcParam(BlendSrc src) {
		int ret = -1;
		switch (src) {
		case ZERO:
			ret = GLES20.GL_ZERO;
			break;
		case COLOR:
			ret = GLES20.GL_SRC_COLOR;
			break;
		case ONE_MINUS_COLOR:
			ret = GLES20.GL_ONE_MINUS_SRC_COLOR;
			break;
		case ALPHA:
			ret = GLES20.GL_SRC_ALPHA;
			break;
		case ONE_MINUS_ALPHA:
			ret = GLES20.GL_ONE_MINUS_SRC_ALPHA;
			break;
		case CONSTANT_COLOR:
			ret = GLES20.GL_CONSTANT_COLOR;
			break;
		case ONE_MINUS_CONSTANT_COLOR:
			ret = GLES20.GL_ONE_MINUS_CONSTANT_COLOR;
			break;
		case CONSTANT_ALPHA:
			ret = GLES20.GL_CONSTANT_ALPHA;
			break;
		case ONE_MINUS_CONSTANT_ALPHA:
			ret = GLES20.GL_ONE_MINUS_CONSTANT_ALPHA;
			break;
		case ALPHA_SATURATE:
			ret = GLES20.GL_SRC_ALPHA_SATURATE;
			break;
		case ONE:
		default:
			ret = GLES20.GL_ONE;
			break;
		}
		return ret;
	}

	private int getBlenDstParam(BlendDst dst) {
		int ret = -1;
		switch (dst) {
		case ZERO:
			ret = GLES20.GL_ZERO;
			break;
		case COLOR:
			ret = GLES20.GL_DST_COLOR;
			break;
		case ALPHA:
			ret = GLES20.GL_DST_ALPHA;
			break;
		case ONE_MINUS_ALPHA:
			ret = GLES20.GL_ONE_MINUS_DST_ALPHA;
			break;
		case ONE_MINUS_COLOR:
			ret = GLES20.GL_ONE_MINUS_DST_COLOR;
			break;
		case CONSTANT_COLOR:
			ret = GLES20.GL_CONSTANT_COLOR;
			break;
		case ONE_MINUS_CONSTANT_COLOR:
			ret = GLES20.GL_ONE_MINUS_CONSTANT_COLOR;
			break;
		case CONSTANT_ALPHA:
			ret = GLES20.GL_CONSTANT_ALPHA;
			break;
		case ONE_MINUS_CONSTANT_ALPHA:
			ret = GLES20.GL_ONE_MINUS_CONSTANT_ALPHA;
			break;
		case ONE:
		default:
			ret = GLES20.GL_ONE;
			break;
		}
		return ret;
	}

	public void setColorMask(boolean r, boolean g, boolean b, boolean a) {
		this.colorMask[0] = r;
		this.colorMask[1] = g;
		this.colorMask[2] = b;
		this.colorMask[3] = a;
	}

	private int getDepthTestParam(DepthTest depthTest) {
		int ret;
		switch (depthTest) {
		case LESS:
			ret = GLES20.GL_LESS;
			break;
		case GREATER:
			ret = GLES20.GL_GREATER;
			break;
		case LEQUAL:
			ret = GLES20.GL_LEQUAL;
			break;
		case GEQUAL:
			ret = GLES20.GL_GEQUAL;
			break;
		case EQUAL:
			ret = GLES20.GL_EQUAL;
			break;
		case NOT_EQUAL:
			ret = GLES20.GL_NOTEQUAL;
			break;
		case ALWAYS:
		default:
			ret = GLES20.GL_ALWAYS;
			break;
		}
		return ret;
	}

	public void setDepthTest(DepthTest depthTest) {
		this.depthTest = depthTest;
	}

	public void setDepthWriting(boolean doDepthWriting) {
		this.doDepthWriting = doDepthWriting;
	}

	/**
	 * Sets-up all the OpenGL parameters such as writing depth or setting back-face
	 * culling.
	 */
	void setupRendering() {
		if (this.doBackfaceCulling) {
			GLES20.glEnable(GLES20.GL_CULL_FACE);
			GLES20.glCullFace(getCullFaceParam(this.cullFace));
		} else {
			GLES20.glDisable(GLES20.GL_CULL_FACE);
		}

		if (this.doBlending) {
			GLES20.glEnable(GLES20.GL_BLEND);
			GLES20.glBlendColor(
					this.constantBlendColor.getR(),
					this.constantBlendColor.getG(),
					this.constantBlendColor.getB(),
					this.constantBlendColor.getA());
			GLES20.glBlendFunc(
					getBlenSrcParam(this.blendSrc),
					getBlenDstParam(this.blendDst));
		} else {
			GLES20.glDisable(GLES20.GL_BLEND);
		}

		GLES20.glColorMask(this.colorMask[0], this.colorMask[1], this.colorMask[2], this.colorMask[3]);

		GLES20.glDepthMask(this.doDepthWriting);
		GLES20.glDepthFunc(getDepthTestParam(this.depthTest));
	}

	public final int useExistingProgram(int programHandle) {
		this.programHandle = programHandle;
		return this.programHandle;
	}

	public final int createProgram() {
		this.programHandle = GLES20.glCreateProgram();
		return this.programHandle;
	}

	public final int getProgramHandle() {
		return this.programHandle;
	}

	public final int attachVertexShader(String shaderCode) {
		return attachShader(GLES20.GL_VERTEX_SHADER, shaderCode);
	}

	public final int attachFragmentShader(String shaderCode) {
		return attachShader(GLES20.GL_FRAGMENT_SHADER, shaderCode);
	}

	public final int attachShader(int shaderType, String shaderCode) {
		final int handle = loadShader(shaderType, shaderCode);
		if (handle != 0) {
			GLES20.glAttachShader(this.programHandle, handle);
		}
		return handle;
	}

	public final int attachShader(int shaderHandle) {
		GLES20.glAttachShader(this.programHandle, shaderHandle);
		return shaderHandle;
	}

	protected static int loadVertexShader(String shaderCode) {
		return loadShader(GLES20.GL_VERTEX_SHADER, shaderCode);
	}

	protected static int loadFragmentShader(String shaderCode) {
		return loadShader(GLES20.GL_FRAGMENT_SHADER, shaderCode);
	}

	/**
	 * Loads source and compiles a vertex or fragment shader.
	 * @param shaderType - GLES20.GL_VERTEX_SHADER or GLES20.GL_FRAGMENT_SHADER.
	 * @param shaderCode - The string that contains the shader code.
	 * @return The handle of the loaded shader.
	 */
	public static int loadShader(int shaderType, String shaderCode) {
	    int shaderHandle = GLES20.glCreateShader(shaderType);
	    GLES20.glShaderSource(shaderHandle, shaderCode);
	    GLES20.glCompileShader(shaderHandle);

        if (!isShaderCompiled(shaderHandle)) {
            Log.e("kan", "Could not compile shader:\n" + getShaderInfoLog(shaderHandle));
            GLES20.glDeleteShader(shaderHandle);
            shaderHandle = 0;
        }
	    return shaderHandle;
	}

	public void detachShader(int shaderHandle) {
		GLES20.glDetachShader(this.programHandle, shaderHandle);
	}

	public static boolean isShaderCompiled(int shaderHandle) {
		final int[] compiled = { -1 };
		GLES20.glGetShaderiv(shaderHandle, GLES20.GL_COMPILE_STATUS, compiled, 0);
		return compiled[0] != 0;
	}

	public static String getShaderInfoLog(int shaderHandle) {
		return GLES20.glGetShaderInfoLog(shaderHandle);
	}

	public final boolean linkProgram() {
		boolean linked = true;
		GLES20.glLinkProgram(this.programHandle);
		if (!isProgramLinked()) {
			Log.e("kan", "Linking failed:\n" + getProgramInfoLog());
			linked = false;
		}
		return linked;
	}

	public boolean isProgramLinked() {
		final int[] linked = { -1 };
		GLES20.glGetProgramiv(this.programHandle, GLES20.GL_LINK_STATUS, linked, 0);
		return linked[0] != 0;
	}

	public String getProgramInfoLog() {
		return GLES20.glGetProgramInfoLog(this.programHandle);
	}

	public void destroyResources() {
		GLES20.glDeleteProgram(this.programHandle);
		this.programHandle = -1;
	}

	protected final int passFloatVertexArray(
			String variableName, int size, int stride, FloatBuffer data) {
		final int handle = GLES20.glGetAttribLocation(this.programHandle, variableName);
		if (handle != -1) {
    		GLES20.glEnableVertexAttribArray(handle);
    		GLES20.glVertexAttribPointer(handle, size, GLES20.GL_FLOAT, false, stride, data);
		}
		return handle;
	}

	final int passUniformInt(String variableName, int value) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform1i(handle, value);
		}
		return handle;
	}

	final int passUniformFloat(String variableName, float value) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform1f(handle, value);
		}
		return handle;
	}

	final int passUniformVec(String variableName, int[] vec, int vecType) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (vecType) {
			case 1:
				GLES20.glUniform1i(handle, vec[0]);
				break;
			case 2:
				GLES20.glUniform2i(handle, vec[0], vec[1]);
				break;
			case 3:
				GLES20.glUniform3i(handle, vec[0], vec[1], vec[2]);
				break;
			case 4:
				GLES20.glUniform4i(handle, vec[0], vec[1], vec[2], vec[3]);
				break;
			}
		}
		return handle;
	}

	final int passUniformVector3(String variableName, Vector3 v) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			GLES20.glUniform3f(handle, v.getX(), v.getY(), v.getZ());
		}
		return handle;
	}

	final int passUniformVec(String variableName, float[] vec, int vecType) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (vecType) {
			case 1:
				GLES20.glUniform1f(handle, vec[0]);
				break;
			case 2:
				GLES20.glUniform2f(handle, vec[0], vec[1]);
				break;
			case 3:
				GLES20.glUniform3f(handle, vec[0], vec[1], vec[2]);
				break;
			case 4:
				GLES20.glUniform4f(handle, vec[0], vec[1], vec[2], vec[3]);
				break;
			}
		}
		return handle;
	}

	final int passUniformMatrix(String variableName, float[] matrix, int matType) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
			switch (matType) {
			case 2:
				GLES20.glUniformMatrix2fv(handle, 1, false, matrix, 0);
				break;
			case 3:
				GLES20.glUniformMatrix3fv(handle, 1, false, matrix, 0);
				break;
			case 4:
				GLES20.glUniformMatrix4fv(handle, 1, false, matrix, 0);
				break;
			}
		}
		return handle;
	}

	final int passUniformTexture(String variableName, int textureNum, Texture texture) {
		return passUniformTextureHandle(variableName, textureNum, texture.getTextureHandle());
	}

	final int passUniformTextureHandle(String variableName, int textureNum, int textureHandle) {
		final int handle = GLES20.glGetUniformLocation(this.programHandle, variableName);
		if (handle != -1) {
		    GLES20.glActiveTexture(Texture.getGLTexNum(textureNum));
		    GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textureHandle);
		    GLES20.glUniform1i(handle, textureNum);
		}
		return handle;
	}
}
